local cql = ClassicQuestLog

-- key bindings
BINDING_HEADER_CLASSIC_QUEST_LOG = "Classic Quest Log"
BINDING_NAME_CLASSIC_QUEST_LOG_TOGGLE = "Classic Quest Log"

local detailMode = "Detail" -- either "Detail", "Lore" or "Options"; use cql:GetMode() or cql:SetMode()

-- events registered while the window is on screen
local events = {
    "QUEST_WATCH_LIST_CHANGED",
    "GROUP_ROSTER_UPDATE",
    "PARTY_MEMBER_ENABLE",
    "PARTY_MEMBER_DISABLE",
    "QUEST_POI_UPDATE",
    "QUEST_WATCH_UPDATE",
    "UNIT_QUEST_LOG_CHANGED",
    "PLAYER_REGEN_DISABLED",
    "PLAYER_REGEN_ENABLED",
}

cql:RegisterEvent("PLAYER_LOGIN")

local firstShown = true -- becomes false after window first shown

function cql:PLAYER_LOGIN()
    tinsert(UISpecialFrames,"ClassicQuestLog")
    self.TitleContainer.TitleText:SetText(QUEST_LOG)

    -- before touching the settings metatable, see if user had Classic Quest Log 2.x and this is a new 3.x
    local hadCQL2 = next(ClassicQuestLogSettings) and not ClassicQuestLogSettings.LastVersion

    -- these events are always registered to capture changes when window is not on screen
    self:RegisterEvent("QUEST_LOG_UPDATE")
    self:RegisterEvent("QUEST_ACCEPTED")
    self:RegisterEvent("LORE_TEXT_UPDATED_CAMPAIGN")

    local v1,v2,v3 = (GetAddOnMetadata("Classic Quest Log","Version") or ""):match("^(%d+)%.(%d+)%.(%d+)")
    local version = (tonumber(v1) or 0)*10000+(tonumber(v2) or 0)*100+(tonumber(v3) or 0)
    local lastVersion = tonumber(cql.settings.LastVersion) or 0

    -- check if this is the first time a new version has run
    if not cql.settings.LastVersion or version > lastVersion then
        -- if this is first time this has run version 3.0.0 or higher (which didn't track this),
        -- reset all settings except ShowMinimapButton
        if not cql.settings.LastVersion then
            local showMinimapButton = cql.settings.ShowMinimapButton
            local minimapPosition = cql.settings.MinimapPosition
            wipe(ClassicQuestLogSettings)
            cql.settings.ShowMinimapButton = showMinimapButton
            cql.settings.MinimapPosition = minimapPosition
            if hadCQL2 then
                cql.Chrome.OptionsButton.HelpPlate:Show()
            end
        end
        cql.settings.LastVersion = version
    end

    -- on login, trim quest history to keep 10 most recent questIDs, no need to keep more
    for i=#cql.settings.QuestHistory-11,1,-1 do
        tremove(cql.settings.QuestHistory,i)
    end

    if not cql.settings.PerCharacterPosition and cql.settings.XPos and cql.settings.YPos and cql.settings.Width and cql.settings.Height then
        self:ClearAllPoints()
        self:SetPoint("BOTTOMLEFT",cql.settings.XPos,cql.settings.YPos)
        self:SetSize(cql.settings.Width,cql.settings.Height)
    else
        self:UpdateSavedPosition()
    end

    self.Chrome:Setup()
    self.Options:Setup()
    self.Log:Setup()
    self.Detail:Setup()
    self.Lore:Setup()
    self.BindButton:Setup()
    self.MicroButton:Setup()
    self.ObjectiveTracker:Setup()
    self.MinimapButton:Setup()

    -- hook SetSelectedQuest so any outside quest selection updates this addon too
    -- at times multiple can run in same frame (like clicking objective tracker) so
    -- this waits a short delay after last event
    hooksecurefunc(C_QuestLog,"SetSelectedQuest",function(questID)
        if self:IsVisible() then
            cql.Log:AddQuestToHistory(questID)
            cql.timer:Start(0.1,self.DelayedUpdate)
        end
    end)

    -- add slash command to toggle window
    SLASH_CLASSICQUESTLOG1 = "/cql"
    SlashCmdList["CLASSICQUESTLOG"] = self.Toggle

    -- broker button
	local ldb = LibStub and LibStub:GetLibrary("LibDataBroker-1.1",true)
    if ldb then
        ldb:NewDataObject("ClassicQuestLog",{ type="launcher", icon="Interface\\AddOns\\Classic Quest Log\\textures\\cqlbutton", tooltiptext="Classic Quest Log", OnClick=cql.Toggle })
	end

end

-- this is always registered and will update the log's quest list (and then UI after)
function cql:QUEST_LOG_UPDATE(...)
    self.Log:UpdateQuestList()
    self:Update()
end

-- when a quest is accepted, log should select that quest
function cql:QUEST_ACCEPTED(questID)
    self.Log:QuestAccepted(questID)
end

-- when entering combat (only registered while window on screen)
function cql:PLAYER_REGEN_DISABLED()
    -- we are not yet in combat when this event fires; so InCombatLockdown is false
    if self.Chrome.MapButton:IsVisible() then
        self.Chrome.MapButton.disabledReason = "combat"
        self.Chrome.MapButton:Disable()
        self.Chrome.MapButton.Text:SetTextColor(0.5,0.5,0.5)
    end
end

-- when leaving combat (only registered while window on screen)
function cql:PLAYER_REGEN_ENABLED()
    if self.Chrome.MapButton:IsVisible() then
        self.Chrome.MapButton:Update()
    end
end

-- registered in overrides if Blizzard_ObjectiveTracker wasn't loaded on login
function cql:ADDON_LOADED(addon)
    if addon=="Blizzard_ObjectiveTracker" then
        cql.ObjectiveTracker:Setup()
    end
end

-- when lore text changes or response from requests for lore text
-- TODO: check if this fires when lore changes on a campaign quest
function cql:LORE_TEXT_UPDATED_CAMPAIGN(campaignID,lore)
    if campaignID and lore then
        cql.Lore:UpdateCampaignLore(campaignID,lore)
    end
end

function cql:MODIFIER_STATE_CHANGED(key,down)
    if MouseIsOver(cql.Log) then
        local focus = GetMouseFocus()
        if (focus.info and (focus.info.questID or focus.info.campaignID)) or focus.campaignID then
            if GameTooltip:IsVisible() then
                GameTooltip:Hide()
            elseif cql.Log.CampaignTooltip:IsVisible() then
                cql.Log.CampaignTooltip:Hide()
            elseif focus.campaignID and focus:GetParent().LoreButtonOnEnter then
                focus:GetParent():LoreButtonOnEnter()
            elseif focus.OnEnter then
                focus:OnEnter()
            end
        end
    end
end

-- call to update the whole UI
function cql:Update()
    self.Chrome:Update()
    self.Log:Update()

    -- updates either Detail, Lore or Options depending on which mode we're in
    if detailMode and self[detailMode] and self[detailMode].Update then
        self[detailMode]:Update()
    end

    self.BindButton:Update()
    self.MicroButton:Update()
    self.Dialog:Update()
end

-- called after a timer:Start() delay, updates the UI
function cql:DelayedUpdate()
    cql:Update()
end

-- when a DelayedUpdate should also scroll to selected quest after update
function cql:DelayedUpdateAndScrollToSelectedQuest()
    cql:Update()
    cql.Log:ScrollToQuestID(cql.Log:GetSelectedQuest())
end

function cql:OnEvent(event,...)
    if self[event] then
        self[event](self,...)
    else -- if event doesn't have its own function, update the UI
        self:Update()
    end
end

function cql:OnShow()
    for _,event in ipairs(events) do
        self:RegisterEvent(event)
    end
    if not cql.settings.ShowTooltips then
        self:RegisterEvent("MODIFIER_STATE_CHANGED")
    end
    self.Log:UpdateQuestList()

    -- if we are first showing the window in a session, go back to recent questID
    -- (sometime between login and this window showing, default may have selected first quest)
    if firstShown then
        firstShown = false
        local recentQuestID = cql.Log:GetRecentQuestID()
        local currentQuestID = cql.Log:GetSelectedQuest()
        if recentQuestID and recentQuestID~=0 and recentQuestID~=currentQuestID then
            C_QuestLog.SetSelectedQuest(recentQuestID)
        end
    end

    self:Update()
    self.MicroButton:SetButtonState("PUSHED",true)
    self.Lore:RequestCampaignLore()
    PlaySound(SOUNDKIT.IG_QUEST_LOG_OPEN)

    self.timer:Start(0.1,cql.Log.DelayedScrollToSelectedQuestID)
end

function cql:OnHide()
    for _,event in ipairs(events) do
        self:UnregisterEvent(event)
    end
    self:UnregisterEvent("MODIFIER_STATE_CHANGED")
    self:SetMode("Detail",true)
    self.Dialog:Hide()
    self.MicroButton:SetButtonState("NORMAL")
    PlaySound(SOUNDKIT.IG_QUEST_LOG_CLOSE)
end

function cql:OnMouseDown()
    if not self.settings.LockWindow or IsShiftKeyDown() then
        self:StartMoving()
        self.isMoving = true
    end
end

function cql:OnMouseUp()
    if self.isMoving then
        self.isMoving = nil
        self:StopMovingOrSizing()
        self:UpdateSavedPosition()
    end
end

-- toggles the window (don't use self reference in here since toc global can use this)
function cql:Toggle()
    if not cql:IsVisible() then
        cql:Show()
    else
        cql:Hide()
    end
end
ClassicQuestLogToggleWindow = cql.Toggle

-- sets mode to either "Detail", "Lore" or "Options" and moves to that window if not already there
-- set noUpdate to show/hide panel but not update the UI afterwards (such as going to Detail on hide)
function cql:SetMode(mode,noUpdate)
    self.Chrome.OptionsButton.HelpPlate:Hide()
    if mode=="Detail" or mode=="Lore" or mode=="Options" then
        if mode~=detalMode then
            if self[detailMode] then
                self[detailMode]:Hide()
            end
            if self[mode] then
                self[mode]:Show()
            end
            detailMode = mode
            if not noUpdate then
                self:Update()
            end
        end
    end
end

-- returns current mode, either "Detail", "Lore" or "Options"
function cql:GetMode()
    return detailMode
end

-- updates the XPos, YPos, Width and Height savedvars of window position/size
function cql:UpdateSavedPosition()
    cql.settings.XPos = self:GetLeft()
    cql.settings.YPos = self:GetBottom()
    cql.settings.Width = self:GetWidth()
    cql.settings.Height = self:GetHeight()
end
